#![feature(coroutines)]
#![feature(coroutine_trait)]
#![feature(stmt_expr_attributes)]
#![feature(unboxed_closures)]
//#![cfg_attr(feature = "async", feature(async_fn_traits))]

use godot::{classes::Engine, prelude::*};
use godot_tokio::AsyncRuntime;
use pipewire::Pipewire;

mod pipewire;
mod player;
mod portals;

struct Extension;

#[gdextension]
unsafe impl ExtensionLibrary for Extension {
    fn on_level_init(level: InitLevel) {
        #[cfg(not(target_os = "linux"))]
        panic!("Not supported on Windows");
        if level == InitLevel::Scene {
            let mut engine = Engine::singleton();

            // This is where we register our async runtime singleton.
            engine.register_singleton(AsyncRuntime::SINGLETON, &AsyncRuntime::new_alloc());
            engine.register_singleton(Pipewire::SINGLETON, &Pipewire::new_alloc());
        }
    }
    fn on_level_deinit(level: InitLevel) {
        if level == InitLevel::Scene {
            let mut engine = Engine::singleton();

            // Here is where we free our async runtime singleton from memory.
            Extension::free_singleton(&mut engine, AsyncRuntime::SINGLETON);
            Extension::free_singleton(&mut engine, Pipewire::SINGLETON);
        }
    }
}

impl Extension {
    fn free_singleton(engine: &mut Engine, name: &str) {
        if let Some(async_singleton) = engine.get_singleton(name) {
            engine.unregister_singleton(name);
            async_singleton.free();
        } else {
            godot_warn!("Failed to find & free singleton -> {}", name);
        }
    }
}
